"""Implementation of Cycling Performance Metrics
Author: Maksym Sladkov
"""

import numpy as np
from vmpy.preprocess import rolling_mean


POWER_ZONES_THRESHOLD = [0.55, 0.75, 0.90, 1.05, 1.20, 1.50]


def best_interval(stream, window, moving=None, **kwargs):
    """Calculate best interval of the stream

    :param stream: array-like
    :param window: int, duration of the interval in seconds
    :return best_interval: float
    """

    _stream = rolling_mean(stream, window=window, moving=moving, **kwargs)

    if type(_stream) == list:
        _stream = np.asarray(_stream)

    rv = _stream.max()

    return rv


def normalized_power(stream, moving=None, **kwargs):
    """Normalized power

    :param stream: array-like power stream
    :param moving (optional): array-like moving bools
    :param type (optional): str, default='NP', "xPower"
    """

    if kwargs.get('type', 'NP') == 'xPower':
        _rolling_mean = rolling_mean(stream, window=25, moving=moving, type='emwa')
    else:
        _rolling_mean = rolling_mean(stream, window=30, moving=moving)

    if type(_rolling_mean) == list:
        _rolling_mean = np.asarray(_rolling_mean, dtype=np.float)

    rv = np.mean(np.power(_rolling_mean, 4)) ** (1.0 / 4)

    return rv


def relative_intensity(norm_power, threshold_power):
    """Relative intensity

    :param norm_power: NP or xPower
    :param threshold_power: FTP or CP
    :return float: IF or RI
    """

    return norm_power/threshold_power


def stress_score(norm_power, threshold_power, duration):
    """Stress Score

    :param norm_power: NP or xPower
    :param threshold_power: FTP or CP
    :param duration: in seconds
    :return ss: TSS or BikeScore
    """

    ri = relative_intensity(norm_power, threshold_power)

    ss = (duration * norm_power * ri) \
         / (threshold_power * 3600) * 100

    return ss



"""Very thin wrapper around the Strava API v3

    STRAVA_ACCESS_TOKEN must be explicitly provided as an input

    ALL returned values are python objects e.g. dict or list

    Author: Maksym Sladkov
"""
import requests
import json
import logging

logger = logging.getLogger(__name__)


def retrieve_athlete(access_token):
    """Retrieve current(authenticated) athlete
    API V3: https://strava.github.io/api/v3/athlete/#get-details

    :param access_token: Settings/My API Applications/Your Access Token -> str
    :return athlete: -> dict
    """

    endpoint_url = "https://www.strava.com/api/v3/athlete"

    r = requests.get(endpoint_url,
                     headers=authorization_header(access_token))

    if r.ok:

        athlete = json.loads(r.text)

    else:

        logger.error('Retrieve Athlete Failed with a reason {}'.format(r.reason))
        athlete = None

    return athlete


def retrieve_activity(activity_id, access_token):
    """Retrieve a detailed representation of activity
    API V3: https://strava.github.io/api/v3/activities/#get-details

    :param activity_id: Activity ID -> int
    :param access_token: Settings/My API Applications/Your Access Token -> str
    :return activity: -> dict
    """

    endpoint_url = "https://www.strava.com/api/v3/activities/{}".format(activity_id)

    r = requests.get(endpoint_url,
                     headers=authorization_header(access_token))

    if r.ok:

        activity = json.loads(r.text)

    else:

        logger.error('Retrieve Activity Failed with a reason {}'.format(r.reason))
        activity = None

    return activity


def retrieve_streams(activity_id, access_token, **kwargs):
    """Retrieve activity streams
        API V3: https://strava.github.io/api/v3/streams/#activity

        :param activity_id: Activity ID -> int
        :param access_token: Settings/My API Applications/Your Access Token -> str
        :return streams: -> list
        """

    types = kwargs.get("types",
                       "time,latlng,distance,altitude,velocity_smooth,heartrate,cadence,watts,temp,moving,grade_smooth")

    endpoint_url = "https://www.strava.com/api/v3/activities/{}/streams/{}".format(activity_id, types)

    r = requests.get(endpoint_url,
                         headers=authorization_header(access_token))

    if r.ok:

        streams = json.loads(r.text)

    else:

        logger.error('Retrieve Streams Failed with a reason {}'.format(r.reason))
        streams = None

    return streams


def stream2dict(stream_list):
    """Convert stream list into stream dict

    :param stream_list: stream in list form, typical Strava API v3 response
    :return stream_dict: stream in dict form, ready to be consumed by pandas
    """

    stream_dict = {}

    for s in stream_list:

        stream_dict.update({s['type']: s['data']})

    return stream_dict


def authorization_header(access_token):
    """Authorization header dict to be used with requests.get()

    :param access_token: Settings/My API Applications/Your Access Token -> str
    :return header: -> dict
    """

    rv = {'Authorization': 'Bearer {}'.format(access_token)}

    return rv


"""Implementation of help functions for data preprocessing
Author: Maksym Sladkov"""

import numpy as np
import pandas as pd


def sanitize_stream(stream, moving=None, **kwargs):
    """Fill moving=False and remove outliers

    :param stream: array-like
    :param moving (optional): array-like of bools
    :param value (optional): value to use to fill moving=False, default=0.0
    """

    # Infer stream data type and cast to dnarray
    _stream_type = type(stream)
    if _stream_type == list:
        _stream = np.asarray(stream)
    else:
        _stream = stream

    if moving is not None:
        if type(moving) == list:
            _moving = np.asarray(moving, dtype=bool)
        else:
            _moving = moving
    else:
        _moving = np.ones(_stream.size, dtype=bool)

    _stream[~_moving] = kwargs.get('value', 0.0)

    return _stream


def rolling_mean(stream, window, moving=None, **kwargs):
    """Rolling mean (default=10 sec) of the stream

    :param stream: array-like
    :param window: int, Size of the moving window in sec
    :param moving (optional): array-like of boolean to mark samples to use
    :param value (optional): Value to use to fill moving=False, default=0.0
    :param type (optional): str, type of averaging default="uniform", "ewma",
    window is used as a span
    :return y: type of input argument

    The stream is expected to be sampled with 1 sec interval
    and is treated as a time series.

    The moving array will indicate which samples to set to zero before
    applying rolling mean.
    """

    # Set all records with Moving=False to zero
    _stream = sanitize_stream(stream, moving, **kwargs)

    # Moving average using pandas rolling_mean
    _s = pd.Series(_stream)

    if kwargs.get('type', 'uniform') == 'emwa':
        y = _s.ewm(span=window, min_periods=1).mean().values
    else:
        y = _s.rolling(window, min_periods=1).mean().values

    # Cast the result into original type
    if type(stream) == list:
        y = y.tolist()

    return y